
// This master header file was added by D. Lemire in April 2021 to ease integration
// in simdutf.
#ifndef UTF8LUT_H
#define UTF8LUT_H
#include "message/MessageConverter.h"
#include "message/MessageConverter.cpp"

#include "buffer/ProcessorSelector.h"
#include "buffer/BaseBufferProcessor.cpp"

#include "core/DecoderLut.cpp"
#include "core/EncoderLut.cpp"

#endif
